from .emitter import Emitter
from functools import reduce
import logging
import boto3
import json


logger = logging.getLogger()

s3 = boto3.resource('s3')

def handle(event, ctx, reducerFunction):
    logger.debug('got event{}'.format(event))

    em = Emitter(
        ctx.client_context.custom['outputBucket'],
        ctx.client_context.custom['outputPrefix'],
        ctx.client_context.custom['jobId'],
        "reducer"
    )

    reducerKey = event['key']
    outputBucket = ctx.client_context.custom['outputBucket']

    # TODO use parallel fetching
    def getData(path):
        obj = s3.Object(outputBucket, path)
        return json.loads(obj.get()['Body'].read().decode('utf-8'))

    # get items for this partition
    allData = list(map(getData, event['value']))
    logger.debug(f"allData: {allData}")

    # decode data
    def decodeData(allValues, jsonData):
        values = jsonData['values']
        for v in values:
            if(v['valueIsBase64']):
                allValues.append(codecs.decode(v['value'], 'base64'))
            else:
                allValues.append(v['value'])
        return allValues

    values = reduce(decodeData, allData, [])

    # decode key
    key = allData[0]['key']
    if(allData[0]['keyIsBase64']):
        key = codecs.decode(key, 'base64')

    # run reducer
    reducerResult = reducerFunction(key, values, em)
    logger.debug('reducerResult{}'.format(reducerResult))
    em.flushEmits()

    return {
        'reducerResult': reducerResult
    }



from .emitter import Emitter
import logging
import boto3

logger = logging.getLogger()

s3 = boto3.resource('s3')

def handle(event, ctx, mapperFunction):
    logger.debug('got event{}'.format(event))

    em = Emitter(
        ctx.client_context.custom['outputBucket'],
        ctx.client_context.custom['outputPrefix'],
        ctx.client_context.custom['jobId'],
        "mapper"
    )

    mapKey = event['key']
    logger.debug(f"mapKey:{mapKey}")
    inputBucket = ctx.client_context.custom['inputBucket']
    logger.debug(f"inputBucket:{inputBucket}")

    obj = s3.Object(inputBucket, mapKey)
    dataObj = obj.get()['Body'].read()
    logger.debug('dataObjType{}'.format(dataObj))
    mapperResult = mapperFunction(mapKey, dataObj, em)
    logger.debug('mapperResult{}'.format(mapperResult))
    em.flushEmits()

    return {
        'mapperResult': mapperResult
    }


import os
import boto3
import logging
import json
import codecs
import hashlib
import binascii
import inspect
from collections import defaultdict

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

s3 = boto3.resource('s3')

class Emitter(object):

    def __init__(self, outputBucket, outputPrefix, jobId, emitterType):
        self.emitBuffer = defaultdict(list)
        self.outputBucket = outputBucket
        self.outputPrefix = outputPrefix
        self.jobId = jobId
        self.emitterType = emitterType

    def emit(self, key, value):
        self.emitBuffer[key].append(value)

    def flushEmits(self):
        for key in self.emitBuffer:
            self.flushEmit(key, self.emitBuffer[key])

    def flushEmit(self, key, values):
        logger.debug(f"flushing {key}, {values}")
        keyIsBuffer = type(key).__name__ is 'bytes'
        outputKey = codecs.encode(key, 'base64') if keyIsBuffer else key

        values = []
        for v in values:
            valueIsBuffer = type(v).__name__ is 'bytes'
            outputValue = codecs.encode(v, 'base64') if valueIsBuffer else v
            value = {
                'value': outputValue,
                'valueIsBase64': valueIsBuffer,
            }
            values.append(value)

        # for S3 keys
        h = hashlib.sha256()
        if(keyIsBuffer):
            h.update(key)
        else:
            h.update(key.encode('utf-8'))
        partitionKey = h.hexdigest()
        partKey = str(binascii.hexlify(os.urandom(16)), 'ascii')

        flushKey = f"{self.outputPrefix}/{self.jobId}/{self.emitterType}/{partitionKey}/{partKey}"

        logger.debug(f"Flushing key: {key}, values: {values}, to s3://{self.outputBucket}/{flushKey}")

        body = {
          'key': outputKey,
          'keyIsBase64': keyIsBuffer,
          'values': values
        }
        obj = s3.Object(self.outputBucket, flushKey)
        obj.put(Body=json.dumps(body).encode('utf-8'))



from setuptools import setup
setup(
  name = 'cfmr',
  packages = ['cfmr'],
  version = '0.0.12',
  description = 'Python support library for CFMR',
  author = 'Donnie Flood',
  author_email = 'pypi@floodfx.com',
  url = 'https://github.com/floodfx/cfmr-python',
  download_url = 'https://github.com/floodfx/cfmr-python/archive/0.0.8.tar.gz',
  keywords = ['map', 'reduce', 'cloud', 'function', 'aws', 'lambda', 'azure', 'google', 'cloud'],
  classifiers = [],
)
